import Flutter
import UIKit
import VdoFramework

public class SwiftVdocipherFlutterPlugin: NSObject, FlutterPlugin {
    
    private var flutterMethodChannel: FlutterMethodChannel?
    private var flutterOptionDownloaderChannel: FlutterMethodChannel?
    
    public static func register(with registrar: FlutterPluginRegistrar) {
        let channel = FlutterMethodChannel(name: "com.vdocipher.flutter", binaryMessenger: registrar.messenger())
        let instance = SwiftVdocipherFlutterPlugin()
        registrar.addMethodCallDelegate(instance, channel: channel)
        
        let factory = VdoNativeViewFactory(messenger: registrar.messenger())
        registrar.register(factory, withId: "com.vdocipher.flutter/vdoplayerview")
        
        instance.setupMethodChannel(registrar: registrar)
        VdoCipher.logLevel = .info
    }
    
    private func setupMethodChannel(registrar : FlutterPluginRegistrar)
    {
        flutterMethodChannel = FlutterMethodChannel(name: "com.vdocipher.flutter.video_downloader", binaryMessenger: registrar.messenger())
        registrar.addMethodCallDelegate(self, channel: flutterMethodChannel!)
        
        flutterOptionDownloaderChannel = FlutterMethodChannel(name: "com.vdocipher.flutter.option_downloader", binaryMessenger: registrar.messenger())
        
        registrar.addMethodCallDelegate(self, channel: flutterOptionDownloaderChannel!)
    }
    
    public func handle(_ call: FlutterMethodCall, result: @escaping FlutterResult) {
        switch(call.method) {
        case "getNativeIOSLibraryVersion":
            result("iOS Test" + UIDevice.current.systemVersion);
            break;
        case "downloadOptionsWithOtp":
            // flutterOptionDownloaderChannel
            guard let arguments = call.arguments as? [String: Any] else {
                return
            }
            guard let videoId = arguments["videoId"] as? String else { return }
            let customPlayerId = arguments["customPlayerId"] as? String
            
            var downloadOptions: [String: Any] = [:]
            VdoAsset.createAsset(videoId: videoId, playerId: customPlayerId) { asset, error in
                
                guard let asset = asset else {
                    return
                    
                }
                let mediaInfoMap = self.createMediaInfoMap(with: asset)
                
                var audioTrackMap : [String: Any] = [:]
                audioTrackMap["id"] = 0
                audioTrackMap["type"] = "audio"
                audioTrackMap["language"] = ""
                audioTrackMap["bitrate"] = -1
                
                
                var availableTracks : [[String: Any]] = [[String: Any]]()
                availableTracks.append(audioTrackMap)
                var count = 1
                for quality in asset.getVideoQualities() {
                    var videoTrackMap = [String: Any]()
                    
                    videoTrackMap["id"] = count
                    videoTrackMap["type"] = "video"
                    videoTrackMap["width"] = -1
                    videoTrackMap["height"] = -1
                    videoTrackMap["bitrate"] = quality.bitrate
                    count+=1
                    availableTracks.append(videoTrackMap)
                }
                
                downloadOptions["mediaInfo"] = mediaInfoMap
                downloadOptions["mediaId"] = videoId
                downloadOptions["availableTracks"] = availableTracks
                
                self.flutterOptionDownloaderChannel?.invokeMethod("onOptionsReceived", arguments: ["downloadOptions" : downloadOptions])
            }
            
            
            break;
        case "query":
            guard let arguments = call.arguments as? [String: Any] else {
                return
            }
            let mediaIdsRaw = arguments["mediaIds"] as? [String]
            let statusNames = arguments["statusNames"] as? [String]
            
            var mediaIds : [String] = []
            if let mediaIdsRaw = mediaIdsRaw {
                mediaIds = mediaIdsRaw
            }
            
            let downloadStates: [VdoAsset.DownloadState] = getDownloadStateList(downloadStatesRaw: statusNames)
            
            let downloadFilter : DownloadInfoFilter
            if downloadStates.count != 0 {
                downloadFilter = DownloadInfoFilter (videoId: mediaIds.count != 0 ? mediaIds : nil, downloadState: downloadStates)
            }
            else{
                downloadFilter = DownloadInfoFilter (videoId: mediaIds.count != 0 ? mediaIds : nil)
            }
            let downloadedAssests  = VdoCipher.getListOfDownloads(filterBy: downloadFilter).map { self.createOfflineVdoAssetMap(with: $0) }
            result(downloadedAssests)
            break;
        case "addEventListener":
            // listen to the download events
            NotificationCenter.default.addObserver(
                self,
                selector: #selector(handleDownloadProgress(_:)),
                name: .AssetDownloadProgress, object: nil
            )
            
            NotificationCenter.default.addObserver(
                self,
                selector: #selector(handleDownloadStateChange(_:)),
                name: .AssetDownloadStateChanged, object: nil
            )
            
            NotificationCenter.default.addObserver(
                self,
                selector: #selector(handleDownloadFailed(_:)),
                name: .AssetDownloadFailed, object: nil
            )
            break;
        case "removeEventListener":
            // remove listeners of the download events
            NotificationCenter.default.removeObserver(
                self,
                name: .AssetDownloadProgress, object: nil
            )
            
            NotificationCenter.default.removeObserver(
                self,
                name: .AssetDownloadStateChanged, object: nil
            )
            
            NotificationCenter.default.removeObserver(
                self,
                name: .AssetDownloadFailed, object: nil
            )
            break;
        case "enqueue":
            guard let arguments = call.arguments as? [String: Any], let arguments = arguments["arguments"] as? [String: Any], let downloadSelections = arguments["downloadSelections"] as? [String: Any], let downloadOptions = downloadSelections["downloadOptions"] as? [String: Any] else {
                return
            }
            
            guard let otp = downloadOptions["otp"] as? String else { return }
            guard let playbackInfo = downloadOptions["playbackInfo"] as? String else { return }
            guard let videoId = downloadOptions["mediaId"] as? String else { return }
            VdoAsset.createAsset(videoId: videoId) { asset, error in
                if let selectedVideoTrackId = downloadSelections["selectedVideoTrackId"] as? Int, let selectedQuality = asset?.getVideoQualities()[selectedVideoTrackId-1] {
                    asset?.startDownload(otp: otp, playbackInfo: playbackInfo, videoQuality: selectedQuality)
                } else {
                    asset?.startDownload(otp: otp, playbackInfo: playbackInfo)
                }
            }
            
            break;
        case "remove":
            if let videoId: String = call.arguments as? String {
                VdoAsset.createAsset(videoId: videoId) { asset, error in
                    if asset?.downloadState == .downloading || asset?.downloadState == .paused {
                        asset?.cancelDownload()
                    }
                    else {
                        asset?.deleteDownload()
                    }
                }
            }
            else{
                return
            }
            break;
        case "stopDownload":
            if let videoId: String = call.arguments as? String {
                VdoAsset.createAsset(videoId: videoId) { asset, error in
                    asset?.pauseDownload()
                }
            }
            else{
                return
            }
            break;
        case "resumeDownload":
            if let videoId: String = call.arguments as? String {
                VdoAsset.createAsset(videoId: videoId) { asset, error in
                    asset?.resumeDownload()
                }
            }
            else{
                return
            }
            break;
        default:
            result("unsupported call method: " + call.method);
        }
    }
    
    // MARK: Notification handlers
    
    @objc func handleDownloadStateChange(_ notification: Notification) {
        let notice = notification.userInfo!
        let videoId = notice[VdoAsset.Keys.videoId] as? String
        guard let stateRawValue = notice[VdoAsset.Keys.downloadState] as? String,
              let videoId = videoId
        else {
            return
        }
        VdoAsset.createAsset(videoId: videoId) { asset, error in
            if asset != nil {
                asset?.downloadState = VdoAsset.DownloadState(rawValue: (stateRawValue))!
                let assetMap = self.createVdoAssetMap(with: asset!)
                if(asset?.downloadState == .notDownloaded)
                {
                    self.flutterMethodChannel?.invokeMethod("onDeleted", arguments: ["downloadStatus" : assetMap, "mediaId": videoId])
                }
                else if(asset?.downloadState == .downloaded)
                {
                    self.flutterMethodChannel?.invokeMethod("onCompleted", arguments: ["downloadStatus" : assetMap, "mediaId": videoId])
                }
                else {
                    self.flutterMethodChannel?.invokeMethod("onChanged", arguments: ["downloadStatus" : assetMap, "mediaId": videoId])
                }
            }
            
        }
    }
    
    @objc func handleDownloadProgress(_ notification: Notification) {
        let notice = notification.userInfo!
        let videoId = notice[VdoAsset.Keys.videoId] as? String
        guard let percent = notice[VdoAsset.Keys.percentDownloaded] as? Double, let videoId = videoId else {
            return
        }
        VdoAsset.createAsset(videoId: videoId) { asset, error in
            if asset != nil {
                asset?.percentDownload = percent
                let assetMap = self.createVdoAssetMap(with: asset!)
                self.flutterMethodChannel?.invokeMethod("onChanged", arguments: ["downloadStatus" : assetMap, "mediaId": videoId])
            }
            
        }
    }
    
    @objc func handleDownloadFailed(_ notification: Notification) {
        let notice = notification.userInfo!
        let videoId = notice[VdoAsset.Keys.videoId] as? String
        if let message = notice[VdoAsset.Keys.message], let code = notice[VdoAsset.Keys.code], let videoId = videoId
        {
            VdoAsset.createAsset(videoId: videoId) { asset, error in
                if asset != nil {
                    var assetMap = self.createVdoAssetMap(with: asset!)
                    assetMap["reasonDescription"] = message
                    assetMap["reason"] = code
                    self.flutterMethodChannel?.invokeMethod("onFailed", arguments: ["downloadStatus" : assetMap, "mediaId": videoId])
                }
                
            }
        }
    }
    
    private func getDownloadStateList(downloadStatesRaw : [String]?) -> [VdoAsset.DownloadState]{
        var downloadStates : [VdoAsset.DownloadState] = []
        if let downloadStatesRaw = downloadStatesRaw {
            for downloadState in downloadStatesRaw {
                switch(downloadState){
                case "downloaded":
                    downloadStates.append(.downloaded)
                    break;
                case "not downloaded":
                    downloadStates.append(.notDownloaded)
                    break;
                case "paused":
                    downloadStates.append(.paused)
                    break;
                case "downloading":
                    downloadStates.append(.downloading)
                    break;
                default:
                    break;
                }
            }
        }
        return downloadStates
    }
    
    private func createMediaInfoMap(with asset: VdoAsset) -> [String: Any] {
        var mediaInfoMap : [String: Any] = [:]
        
        mediaInfoMap["title"] = asset.title
        mediaInfoMap["duration"] = Int(asset.duration ?? 0) * 1000
        mediaInfoMap["mediaId"] = asset.videoId
        mediaInfoMap["description"] = ""
        
        return mediaInfoMap
    }
    
    private func createMediaInfoMap(with mediaInfo: MediaInfo) -> [String: Any] {
        var mediaInfoMap : [String: Any] = [:]
        
        mediaInfoMap["title"] = mediaInfo.title
        mediaInfoMap["duration"] = Int(mediaInfo.duration ) * 1000
        mediaInfoMap["mediaId"] = mediaInfo.videoId
        mediaInfoMap["description"] = ""
        
        return mediaInfoMap
    }
    
    private func createVdoAssetMap(with asset: VdoAsset) -> [String: Any] {
        var downloadStatus: [String: Any] = [:]
        
        downloadStatus["mediaInfo"] = createMediaInfoMap(with: asset)
        downloadStatus["videoId"] = asset.videoId
        downloadStatus["percentDownloaded"] = Int(asset.percentDownload * 100)
        downloadStatus["downloadState"] = asset.downloadState.rawValue
        downloadStatus["aspectRatio"] = asset.aspectRatio
        if (asset.posters?.isEmpty == false){
            downloadStatus["poster"] = asset.posters?[0].url
        }
        else{
            downloadStatus["poster"] = ""
        }
        return downloadStatus
    }
    
    private func createOfflineVdoAssetMap(with downloadInfo: DownloadInfo) -> [String: Any] {
        var downloadStatus: [String: Any] = [:]
        
        downloadStatus["mediaInfo"] = createMediaInfoMap(with: downloadInfo.mediaInfo)
        downloadStatus["videoId"] = downloadInfo.mediaInfo.videoId
        downloadStatus["percentDownloaded"] = Int(downloadInfo.percentDownloaded * 100)
        downloadStatus["downloadState"] = downloadInfo.status.rawValue
        downloadStatus["aspectRatio"] = 1
        downloadStatus["poster"] = downloadInfo.poster;
        return downloadStatus
    }
}
