//
//  VdoPlayerPlatformView.swift
//  vdocipher_flutter
//
//  Created by Manish Gupta on 10/01/22.
//

import AVKit
import Flutter
import VdoFramework
import MediaPlayer

class VdoNativeView: NSObject, FlutterPlatformView {
    private var _view: ObservableView
    private var methodChannel : FlutterMethodChannel?
    private var eventChannel : FlutterEventChannel?
    private var asset: VdoAsset?
    private var player: AVPlayer?
    private var targetPlaybackSpeed: Float = 1.0
    private var playerLayer : AVPlayerLayer?
    private var loadResult : FlutterResult?
    private var playerEventSink: FlutterEventSink?
    private var playWhenReady: Bool = true
    private var playerState: PlayerState = .unknown {
        didSet{
            guard let eventChannelSink = playerEventSink else {return}
            let event = ["name" : "onPlayerStateChanged", "playWhenReady" : self.playWhenReady, "playbackState": self.playerState.rawValue] as [String : Any]
            eventChannelSink(event)
        }
    }
    private var playerItemBufferEmptyObserver: NSKeyValueObservation?
    private var playerItemBufferKeepUpObserver: NSKeyValueObservation?
    private var playerItemBufferFullObserver: NSKeyValueObservation?

    private var shouldBuildWithControls: Bool = false
    var parentController: FlutterViewController?
    private lazy var flutterEngineWindow: UIWindow? = (UIApplication.shared.delegate as? FlutterAppDelegate)?.window

    private var availableTracks: [[String: Any]] = [[String: Any]]()
    private var selectedTracks: [[String: Any]] = [[String: Any]]()
    
    init(
        frame: CGRect,
        viewIdentifier viewId: Int64,
        arguments args: Any?,
        binaryMessenger messenger: FlutterBinaryMessenger?
    ) {
        _view = ObservableView()
        super.init()
        VdoCipher.setPlaybackDelegate(delegate: self)

        if let args = args as? [String: Any],
           let shouldBuildWithControls = args["buildWithControls"] as? Bool,
           shouldBuildWithControls {
            self.shouldBuildWithControls = shouldBuildWithControls
        }

        if shouldBuildWithControls {
            VdoCipher.getVdoPlayerViewController().loadViewIfNeeded()
            parentController = flutterEngineWindow?.rootViewController as? FlutterViewController
            VdoCipher.getVdoPlayerViewController().view.frame = _view.bounds
            VdoCipher.getVdoPlayerViewController().view.layer.frame = _view.bounds
            parentController?.addChild(VdoCipher.getVdoPlayerViewController())
            _view.addSubview(VdoCipher.getVdoPlayerViewController().view)
            VdoCipher.getVdoPlayerViewController().didMove(toParent: parentController!)
            VdoCipher.setUIPlayerDelegate(self)
            let view = VdoCipher.getVdoPlayerViewController().view
            view?.setNeedsLayout()
            view?.layoutIfNeeded()
        }

        _view.handler = { [weak self] needsLayout in
            guard let self = self else {return}
            if needsLayout == true
            {
                self.playerLayer?.frame = self._view.bounds
                self.playerLayer?.setNeedsLayout()
            }
        }

        guard let messenger = messenger else {
            return
        }
        methodChannel = FlutterMethodChannel.init(name: "com.vdocipher.flutter.vdoplayerview_\(viewId)",
                                                  binaryMessenger: messenger)
        let eventChannel = FlutterEventChannel.init(name: "com.vdocipher.flutter.vdoplayerview.events_\(viewId)",
                                                    binaryMessenger: messenger)
        eventChannel.setStreamHandler(EventChannelStreamHandler.sharedInstance)
        methodChannel?.setMethodCallHandler({ [weak self]

            (call: FlutterMethodCall, result: @escaping FlutterResult) -> Void in
            guard let self = self else {return}

            switch (call.method) {
            case "load":
                do{
                    guard let arguments = call.arguments as? [String: Any] else {
                        return
                    }
                    let restore = arguments["restore"] as? Bool
                    let embedInfo = arguments["embedInfo"] as? [String: Any]
                    let embedInfoOptions = embedInfo?["embedInfoOptions"] as? [String: Any]
                    let resumeTimeMs = embedInfoOptions?["resumeTimeMs"] as? Int
                    let autoPlay = embedInfoOptions?["autoplay"] as? Bool
                    let preferredCaptionsLanguage = embedInfoOptions?["preferredCaptionsLanguage"] as? String
                    let customPlayerId = embedInfoOptions?["customPlayerId"] as? String
                    self.playWhenReady = autoPlay ?? true
                    let resumeTime = Float(resumeTimeMs ?? 0)
                    if restore == true
                    {
                        if self.player != nil
                        {
                            result(MessageUtils.loadResultMap(asset:self.asset, playWhenReady: self.playWhenReady))
                            PlayerHandler.sharedInstance.shouldResume = false
                            if let lastPlaybackSpeed = arguments["lastPlaybackSpeed"] as? Float {
                                self.targetPlaybackSpeed = lastPlaybackSpeed
                            }
                            self.updatePlayerState()
                            return
                        }
                        guard let asset = PlayerHandler.sharedInstance.asset, let otp = PlayerHandler.sharedInstance.otp, let playbackInfo = PlayerHandler.sharedInstance.playbackInfo else {
                            result("error message: Asset meta-data load failed! code: 3009")
                            return
                        }    //Need to send error back to flutter in case of return
                        self.asset = asset
                        DispatchQueue.main.async {
                            _ = self.setCaptionLanguage(lanCode: preferredCaptionsLanguage)
                        }
                        result(MessageUtils.loadResultMap(asset:self.asset, playWhenReady: self.playWhenReady))
                        DispatchQueue.main.async {
                            if asset.downloadState == .downloaded {
                                asset.playOffline()
                            } else {
                                asset.playOnline(otp: otp, playbackInfo: playbackInfo)
                            }
                        }
                    }
                    else
                    {
                        guard let embedInfo = arguments["embedInfo"] as? [String: Any],
                              let offline = embedInfo["offline"] as? Bool
                        else {
                            return
                        }
                        var videoId : String? = nil
                        var playbackInfo : String? = nil
                        if offline {
                            guard let videoIdRaw = embedInfo["mediaId"] as? String else{
                                return
                            }
                            videoId = videoIdRaw
                        }
                        else{
                            guard let playbackInfoRaw = embedInfo["playbackInfo"] as? String,
                                  let data = Data(base64Encoded: playbackInfoRaw),
                                  let videoInfo = try? JSONSerialization.jsonObject(with: data, options: []) as? [String: Any],
                                  let videoIdRaw = videoInfo["videoId"] as? String else{
                                return
                            }
                            videoId = videoIdRaw
                            playbackInfo = playbackInfoRaw
                        }
                        self.loadResult = result
                        if resumeTime > 0 {
                            PlayerHandler.sharedInstance.resumption = CMTime(value: CMTimeValue(resumeTime), timescale: 1000)
                            PlayerHandler.sharedInstance.shouldResume = true
                            let userDefaults: UserDefaults? = UserDefaults.init(suiteName: "com.vdocipher.ios.VdoFramework")
                            if var assetsLastPlaybackTimeSet = userDefaults?.value(forKey: "assetsLastPlaybackTimeSet") as? [String: Double] {
                                assetsLastPlaybackTimeSet[videoId!] = Double(resumeTime)
                                userDefaults?.set(assetsLastPlaybackTimeSet, forKey: "assetsLastPlaybackTimeSet")
                            }
                        } else {
                            PlayerHandler.sharedInstance.shouldResume = false
                        }
                        VdoAsset.createAsset(videoId: videoId!, playerId: customPlayerId) { [weak self] asset, displayableError  in
                            guard let self = self else {return}
                            if let asset = asset
                            {
                                self.asset = asset
                                DispatchQueue.main.async {
                                    _ = self.setCaptionLanguage(lanCode: preferredCaptionsLanguage)
                                }
                                PlayerHandler.sharedInstance.asset = asset
                                if let loadResult = self.loadResult
                                {
                                    loadResult(MessageUtils.loadResultMap(asset:self.asset, playWhenReady: self.playWhenReady))
                                    self.loadResult = nil
                                }
                                DispatchQueue.main.async {
                                    if asset.downloadState == .downloaded && offline == true {
                                        asset.playOffline()
                                    } else {
                                        guard let otp = embedInfo["otp"] as? String else {
                                            return
                                        }
                                        PlayerHandler.sharedInstance.otp = otp
                                        PlayerHandler.sharedInstance.playbackInfo = playbackInfo
                                        asset.playOnline(otp: otp, playbackInfo: playbackInfo!)

                                    }
                                }
                            }
                            else if let displayableError = displayableError
                            {
//                                if let loadResult = self.loadResult
//                                {
//                                    loadResult(MessageUtils.loadResultMap(asset:self.asset, playWhenReady: false))
//                                    self.loadResult = nil
//                                }
//                              result("error message: " + displayableError.message + " code: \(displayableError.code)")
                                methodChannel?.invokeMethod("onLoadError", arguments: ["errorDescription": MessageUtils.loadErrorInfoMap(vdoError: displayableError)] as [String : Any]);
                                guard let eventChannelSink = EventChannelStreamHandler.sharedInstance.playerEventSink else {return}
                                let event = ["name": "onError", "errorDescription": MessageUtils.loadErrorInfoMap(vdoError: displayableError)] as [String : Any]
                                eventChannelSink(event)
                            }
                        }
                    }
                }
            case "setPlayWhenReady":
                do {
                    if call.arguments as? Int == 0 {
                        self.playWhenReady = false
                        self.player?.pause()
                    } else {
                        self.playWhenReady = true
                        self.player?.rate = self.targetPlaybackSpeed
                    }
                }
            case "getPlayWhenReady":
                do{
                    result(self.playerState == .ready ? self.playWhenReady : false)
                }
            case "getPlaybackAndBufferedPositions":
                do{
                    let currentTime = Int(self.player?.currentItem?.currentTime().seconds.rounded( .up) ?? 0.0 ) * 1000
                    result(["position": currentTime, "bufferedPosition": currentTime])
                }
            case "getDuration":
                do{
                    result(Int(self.player?.currentItem?.duration.seconds.rounded( .up) ?? 0.0 ) * 1000)
                }
            case "seek":
                do{
                    guard let currentTime = call.arguments as? Int else {return}
                    self.player?.seek(to: CMTime(value: CMTimeValue(currentTime), timescale: 1000))
                    result(nil)
                }
            case "stop":
                do{
                    
                }
            case "dispose":
                // Will be called when the player view is about to be disposed.
                do {
                    self.playerLayer?.removeFromSuperlayer()
                    self.playerLayer = nil
                    self.removeObservers()
                    self.player?.replaceCurrentItem(with: nil)
                    self.player = nil
                    DispatchQueue.main.asyncAfter(deadline: (.now() + 0.3)) {
                        MPNowPlayingInfoCenter.default().nowPlayingInfo = nil
                    }
                }
            case "packUp":
                do{
                    PlayerHandler.sharedInstance.resumption = self.player?.currentTime()
                    PlayerHandler.sharedInstance.shouldResume = true
                    self.player?.pause()
                }
            case "setPlaybackSpeed":
                do{
                    guard let speed = call.arguments as? Double else {return}
                    self.targetPlaybackSpeed = Float(speed)
                    if self.player?.rate ?? 0 > 0.0 {
                        self.player?.rate = self.targetPlaybackSpeed
                    }
                }
            case "setAdaptive":
                do{

                }
            case "setAudioTrack":
                do{

                }
            case "setVideoTrack":
                do{

                }
            case "setSubtitleLanguage":
                do{
                    DispatchQueue.main.async {
                        let isSuccess : Bool = self.setCaptionLanguage(lanCode: call.arguments as? String)
                        result(isSuccess)
                    }
                }
            case "getPlaybackProperty":
                do{
                    guard let propertyName = call.arguments as? String else {return}
                    switch(propertyName)
                    {
                    case "totalPlayed":
                        result(VdoCipher.getValueOfPlaybackProperty(VdoCipher.PlaybackProperty.totalPlayed)?["totalPlayed"] ?? 0)
                        break;
                    case "totalCovered":
                        result(VdoCipher.getValueOfPlaybackProperty(VdoCipher.PlaybackProperty.totalCovered)?["totalCovered"] ?? 0)
                        break;
                    default:
                        break;
                    }
                }
            default:
                do{

                }
            }
        })
        methodChannel?.invokeMethod("onPlatformPlayerCreated", arguments: viewId)
    }
    
    private func setCaptionLanguage(lanCode: String?) -> Bool {

        var isSuccess : Bool = false
        if self.asset != nil {
            if lanCode == nil {
                isSuccess = ((self.asset?.setCaption(nil)) != nil)
                return isSuccess;
            }
            for caption in self.asset?.captions ?? [] {
                if caption.lang == lanCode {
                    isSuccess = ((self.asset?.setCaption(caption)) != nil)
                    break;
                }
            }
        }
        return isSuccess
    }

    private func updatePlayerState()
    {
        //Restoring update player state based on player rate.
        if self.player?.rate != 0
        {
            self.playWhenReady = true
        }
        else
        {
            self.playWhenReady = false
        }
        self.playerState = .ready
    }
    
    @objc func handleInterruption(notification: Notification) {
        //        guard let eventChannelSink = playerEventSink else {return}
        //        let event = ["name" : "playbackSessionInterrupted"]
        //        eventChannelSink(event)
        guard let userInfo = notification.userInfo,
              let typeValue = userInfo[AVAudioSessionInterruptionTypeKey] as? UInt,
              let type = AVAudioSession.InterruptionType(rawValue: typeValue) else {
            return
        }
        
        // Switch over the interruption type.
        switch type {
            
        case .began:
            do{
                // An interruption began. Update the UI as necessary.
            }
        case .ended:
            do{
                // An interruption ended. Resume playback, if appropriate.
                guard let optionsValue = userInfo[AVAudioSessionInterruptionOptionKey] as? UInt else { return }
                let options = AVAudioSession.InterruptionOptions(rawValue: optionsValue)
                if options.contains(.shouldResume) {
                    // An interruption ended. Resume playback.
                    self.player?.rate = self.targetPlaybackSpeed
                } else {
                    // An interruption ended. Don't resume playback.
                }
                updatePlayerState()
            }
        default: ()
        }
    }
    
    @objc func handleRouteChange(notification: Notification) {
        guard let userInfo = notification.userInfo,
              let reasonValue = userInfo[AVAudioSessionRouteChangeReasonKey] as? UInt,
              let reason = AVAudioSession.RouteChangeReason(rawValue: reasonValue) else {
            return
        }
        
        // Switch over the route change reason.
        switch reason {

        case .newDeviceAvailable: // New device found.
            do{
                
            }

        case .oldDeviceUnavailable: // Old device removed.
            do{
                self.playWhenReady = false
                self.playerState = .ready
            }
            

        default: ()
        }
    }
    
    func addObserversToPlayer()
    {
        NotificationCenter.default.addObserver(self,
                                               selector: #selector(videoDidEnd),
                                               name: NSNotification.Name.AVPlayerItemDidPlayToEndTime,
                                               object: nil)
        
        NotificationCenter.default.addObserver(self,
                                               selector: #selector(handleInterruption),
                                               name: AVAudioSession.interruptionNotification,
                                               object: AVAudioSession.sharedInstance())
        
        NotificationCenter.default.addObserver(self,
                                               selector: #selector(handleRouteChange),
                                               name: AVAudioSession.routeChangeNotification,
                                               object: nil)
        
        playerItemBufferEmptyObserver = self.player?.currentItem?.observe(\AVPlayerItem.isPlaybackBufferEmpty, options: [.new]) { [weak self] (_, _) in
            guard let self = self else { return }
            self.playerState = .buffering
        }
        
        playerItemBufferKeepUpObserver = self.player?.currentItem?.observe(\AVPlayerItem.isPlaybackLikelyToKeepUp, options: [.new]) { [weak self] (_, _) in
            guard let self = self else { return }
            self.playerState = .ready
        }
        
        playerItemBufferFullObserver = self.player?.currentItem?.observe(\AVPlayerItem.isPlaybackBufferFull, options: [.new]) { [weak self] (_, _) in
            guard let self = self else { return }
            self.playerState = .ready
        }
    }
    
    func removeObservers()
    {
        NotificationCenter.default.removeObserver(self,
                                                  name: NSNotification.Name.AVPlayerItemDidPlayToEndTime,
                                                  object: nil)
        NotificationCenter.default.removeObserver(self,
                                                  name: AVAudioSession.interruptionNotification,
                                                  object: AVAudioSession.sharedInstance())
        NotificationCenter.default.removeObserver(self,
                                                  name: AVAudioSession.routeChangeNotification,
                                                  object: nil)
        
        playerItemBufferEmptyObserver?.invalidate()
        playerItemBufferEmptyObserver = nil

        playerItemBufferKeepUpObserver?.invalidate()
        playerItemBufferKeepUpObserver = nil

        playerItemBufferFullObserver?.invalidate()
        playerItemBufferFullObserver = nil
    }
    
    @objc func videoDidEnd(notification: NSNotification) {
        guard let eventChannelSink = playerEventSink else {return}
        let event = ["name" : "onMediaEnded"]
        eventChannelSink(event)
        self.playerState = .ended
    }
    
    func view() -> UIView {
        return _view
    }
    
    func createNativeView(view _view: UIView){
        _view.backgroundColor = UIColor.blue
        let nativeLabel = UILabel()
        nativeLabel.text = "Native text from iOS"
        nativeLabel.textColor = UIColor.white
        nativeLabel.textAlignment = .center
        nativeLabel.frame = CGRect(x: 0, y: 0, width: 180, height: 48.0)
        _view.addSubview(nativeLabel)
    }
    
    deinit{
        removeObservers()
    }
}

extension VdoNativeView : AssetPlaybackDelegate {
    func streamLoadError(error: VdoError) {
        methodChannel?.invokeMethod("onError", arguments: ["errorDescription": MessageUtils.loadErrorInfoMap(vdoError: error)] as [String : Any]);
        guard let eventChannelSink = EventChannelStreamHandler.sharedInstance.playerEventSink else {return}
        let event = ["name":"onError", "errorDescription": MessageUtils.loadErrorInfoMap(vdoError: error)] as [String : Any]
        eventChannelSink(event)
    }
    
    func streamPlaybackManager(playerReadyToPlay player: AVPlayer) {
        // player is ready to play
        self.setPlay()
    }
    
    func streamPlaybackManager(playerCurrentItemDidChange player: AVPlayer) {
        // player current item has changed
        if !shouldBuildWithControls {
            playerLayer?.removeFromSuperlayer()
            self.removeObservers()
            self.player = player
            self.addObserversToPlayer()
            playerLayer = AVPlayerLayer(player: player)
            playerLayer?.frame = _view.bounds
            _view.layer.addSublayer(playerLayer!)
        } else {
            playerLayer?.removeFromSuperlayer()
            self.removeObservers()
            self.player = player
            self.addObserversToPlayer()
        }
        
        self.availableTracks.removeAll()
        self.selectedTracks.removeAll()

        var id = 0
        for caption in asset?.captions ?? [] {
            let lanCode = caption.lang
            availableTracks.append(["id": id, "type": "subtitle", "language": lanCode as Any])
            id+=1
        }
        
        let enabledLanCode = asset?.getCurrentEnabledCaption()?.lang ?? nil
        for availableTrack in availableTracks {
            let lanCode = availableTrack["language"] as? String
            if lanCode == enabledLanCode {
                let selectedTrack = availableTrack
                selectedTracks.append(selectedTrack)
                break;
            }
        }
        
        guard let eventChannelSink = playerEventSink else {return}
        let event = ["name": "onTracksChanged", "available": availableTracks, "selected": selectedTracks] as [String : Any]
        eventChannelSink(event)
    }
    
    func setPlay()
    {
        self.playerState = .ready
        if PlayerHandler.sharedInstance.shouldResume == true {
            if let resumption = PlayerHandler.sharedInstance.resumption {
                PlayerHandler.sharedInstance.shouldResume = false
                player?.seek(to: resumption, completionHandler: { [weak self] _ in
                    self?.player?.rate = self?.targetPlaybackSpeed ?? 1.0
                })
            }
        } else if self.playWhenReady == true {
            player?.rate = self.targetPlaybackSpeed
        }
    }
}

extension VdoNativeView : FlutterStreamHandler{
    func onListen(withArguments arguments: Any?, eventSink events: @escaping FlutterEventSink) -> FlutterError? {
        self.playerEventSink = events
        return nil
    }
    
    func onCancel(withArguments arguments: Any?) -> FlutterError? {
        self.playerEventSink = nil
        return nil
    }
}

private class MessageUtils{
    static func loadResultMap(asset : VdoAsset?, playWhenReady: Bool) -> [String: Any] {
        var result = [String : Any] ()
        if let asset = asset {
            result["didLoad"] = true
            result["mediaInfo"] = MessageUtils.loadMediaInfoMap(asset:asset)
            result["playWhenReady"] = playWhenReady
        }
        else
        {
            result["didLoad"] = false
        }
        return result
    }
    
    static func loadMediaInfoMap(asset : VdoAsset) -> [String: Any]  {
        var mediaInfoMap = [String : Any] ()
        mediaInfoMap["mediaId"] = asset.videoId
        mediaInfoMap["type"] = "fps"
        mediaInfoMap["title"] = asset.title
        mediaInfoMap["description"] = asset.description
        mediaInfoMap["duration"] = ( Int(asset.duration ?? 0.0) * 1000 )
        return mediaInfoMap
    }
    
    static func loadErrorInfoMap(vdoError: VdoError) -> [String: Any] {
        var errorInfoMap = [String : Any] ()
        errorInfoMap["errorCode"] = vdoError.code
        errorInfoMap["errorMsg"] = vdoError.message
        errorInfoMap["httpStatusCode"] = -1
        return errorInfoMap
    }
}

private class ObservableView: UIView{
    
    var handler : ((Bool) -> Void)?
    
    override func layoutSubviews() {
        self.handler?(true)
    }

}

private enum PlayerState : String
{
    case buffering = "buffering",
         ready = "ready",
         ended = "ended",
         unknown = "unknown"
}

private class PlayerHandler : NSObject
{
    static let sharedInstance = PlayerHandler()
    var asset : VdoAsset?
    var otp : String?
    var playbackInfo: String?
    var resumption : CMTime?
    var shouldResume : Bool?
    private override init() {
        super.init()
    }
}

private class EventChannelStreamHandler : NSObject, FlutterStreamHandler {
    static let sharedInstance = EventChannelStreamHandler()
    var playerEventSink: FlutterEventSink?

    override private init() {
        super.init()
    }

    func onListen(withArguments arguments: Any?, eventSink events: @escaping FlutterEventSink) -> FlutterError? {
        self.playerEventSink = events
        return nil
    }

    func onCancel(withArguments arguments: Any?) -> FlutterError? {
        self.playerEventSink = nil
        return nil
    }
}

// MARK:- VdoPlayerViewControllerDelegate - UI Player callbacks

extension VdoNativeView: VdoPlayerViewControllerDelegate {
    func didTapPlay(for id: String) {
        //Update player state
        self.updatePlayerState()
    }
    func didTapPause(for id: String) {
        //Update player state
        self.updatePlayerState()
    }
    func didTapForward(for id: String) {}
    func didTapBackward(for id: String) {}
    func didTapClose(for id: String) {}
    func didTapEnterFullScreen(for id: String) {
        self.methodChannel?.invokeMethod("onFullscreenChange", arguments: ["isFullscreen": true] as [String: Bool])
    }
    func didTapExitFullScreen(for id: String) {
        self.methodChannel?.invokeMethod("onFullscreenChange", arguments: ["isFullscreen": false] as [String: Bool])
    }
    func didTapPip(for id: String) {}
    func didTapAirplay(for id: String) {}
    func didSeek(for id: String, to time: Float64) {
        let event = ["name": "onPlaybackPositionChanged", "value": time] as [String : Any]
        guard let eventChannelSink = EventChannelStreamHandler.sharedInstance.playerEventSink else {return}
        eventChannelSink(event)
    }
    func didChangePlayerRate(for id: String, to speed: Float) {
        self.targetPlaybackSpeed = speed
        self.updatePlayerState()
        if(speed != 0)
        {
            //Only update if player rate is not zero
            let event = ["name": "onPlaybackSpeedChanged", "value": speed] as [String : Any]
            guard let eventChannelSink = EventChannelStreamHandler.sharedInstance.playerEventSink else {return}
            eventChannelSink(event)
        }
    }

    func didChangeCaption(for id: String, to name: String, code: String) {
        if let selectedSubtitle = self.availableTracks.first(where: { track in
            track["language"] as! String == code as String
        })
        {
            self.selectedTracks.removeAll()
            self.selectedTracks.append(selectedSubtitle)
            guard let eventChannelSink = EventChannelStreamHandler.sharedInstance.playerEventSink else {return}
            let event = ["name": "onTracksChanged", "available": self.availableTracks, "selected": self.selectedTracks] as
            [String : Any]
            eventChannelSink(event)
        }
    }
    func didChangeChapter(for id: String, to name: String, at time: Float64) {}

    // Player periodic callback
    func didUpdatePlayback(for id: String, toTime time: Float64) {}
}
