// ignore_for_file: constant_identifier_names

import 'dart:collection';

import 'package:flutter/services.dart';
import 'package:vdocipher_flutter/vdocipher_flutter.dart';

/// Class used to specify filters for a query
class Query {
  List<String> mediaIds = List.empty(growable: true);
  List<int> statuses = List.empty(growable: true);
  final List<String> _statusNames = List.empty(growable: true);

  Query();

  /// Include only media downloads with the given mediaIds
  Query filterByMediaId(List<String>? mediaIds) {
    if (mediaIds != null) this.mediaIds.addAll(mediaIds);
    return this;
  }

  /// Include only media downloads with the given statuses
  Query filterByStatus(List<int>? statuses) {
    this.statuses.clear();
    if (statuses != null) {
      this.statuses.addAll(statuses);
      _generateStatusName();
    }
    return this;
  }

  _generateStatusName() {
    _statusNames.clear();
    for (int status in statuses) {
      switch (status) {
        case VdoDownloadManager.STATUS_COMPLETED:
          _statusNames.add("downloaded");
          break;
        case VdoDownloadManager.STATUS_DOWNLOADING:
          _statusNames.add("downloading");
          break;
        case VdoDownloadManager.STATUS_PAUSED:
          _statusNames.add("paused");
          break;
      }
    }
  }

  HashMap<String, Object> _toMap() {
    HashMap<String, Object> queryMap = HashMap();
    queryMap['mediaIds'] = mediaIds;
    queryMap['statues'] = statuses;
    queryMap['statusNames'] = _statusNames;
    return queryMap;
  }
}

abstract class EventListener {
  /// Called when a new download is added to queue.
  void onQueued(String mediaId, DownloadStatus downloadStatus);

  /// Called when status of a download changes from pending to downloading or if a download
  /// progresses.
  void onChanged(String mediaId, DownloadStatus downloadStatus);

  /// Called when download completes.
  void onCompleted(String mediaId, DownloadStatus downloadStatus);

  /// Called when a download fails.
  void onFailed(String mediaId, DownloadStatus downloadStatus);

  /// Called when a media download is marked as deleted.
  /// Actual deletion of any downloaded media assets is done asynchronously in the background.
  void onDeleted(String mediaId);
}

/// VdoDownloadManager handles enqueuing download requests and management of media downloads.
///
/// New download requests can be added which will remain in queue until the requirements for downloads to progress are met.
///
/// Status of media downloads can be requested using filtered queries.
///
/// Download events can be listened to by registering a listener with the [addDownloadEventListener].
/// Make sure to unregister listener when not required to avoid memory leaks using [removeDownloadEventListener].

class VdoDownloadManager {
  static const int STATUS_NOT_FOUND = 1;
  static const int STATUS_PENDING = 2;
  static const int STATUS_DOWNLOADING = 3;
  static const int STATUS_PAUSED = 4;
  static const int STATUS_COMPLETED = 5;
  static const int STATUS_FAILED = 6;
  static const int STATUS_REMOVING = 7;
  static const int STATUS_RESTARTING = 8;

  bool _isDownloadEventListenerAdded = false;
  final List<EventListener> _downloadEventListeners =
      List.empty(growable: true);

  final Set<String> _currentMediaIds = HashSet();

  static VdoDownloadManager? _vdoDownloadManager;
  static const MethodChannel _methodChannel =
      MethodChannel("com.vdocipher.flutter.video_downloader");

  VdoDownloadManager() {
    _methodChannel
        .setMethodCallHandler((call) => _handlePlatformMethodCall(call));
  }

  /// Get an instance of [VdoDownloadManager]
  static getInstance() {
    _vdoDownloadManager ??= VdoDownloadManager();
    return _vdoDownloadManager;
  }

  /// Enqueue a request to be downloaded
  void enqueue(DownloadRequest downloadRequest) {
    Map<String, Object?> downloadRequestMap =
        DownloadRequest.toMap(downloadRequest);
    _methodChannel.invokeMethod("enqueue", {"arguments": downloadRequestMap});
  }

  /// Cancel media download and delete them locally
  void remove(String mediaId) {
    _methodChannel.invokeMethod("remove", mediaId);
  }

  /// Stop download for given media id
  void stopDownload(String mediaId) {
    _methodChannel.invokeMethod("stopDownload", mediaId);
  }

  /// Resume download for given media id
  void resumeDownload(String mediaId) {
    _methodChannel.invokeMethod("resumeDownload", mediaId);
  }

  /// Query the [VdoDownloadManager] about media downloads that have been requested. Returns future of [DownloadStatus] list.
  Future<List<DownloadStatus>> query(Query query) async {
    List<Object?> mapList =
        await _methodChannel.invokeMethod("query", query._toMap());

    List<DownloadStatus> downloadStatusList = List.empty(growable: true);

    for (Object? downloadStatusObject in mapList) {
      Map<Object?, Object?> downloadStatusMap =
          downloadStatusObject as Map<Object?, Object?>;
      DownloadStatus downloadStatus = DownloadStatus.fromMap(downloadStatusMap);
      _currentMediaIds.add(downloadStatus.mediaInfo.mediaId);
      downloadStatusList.add(downloadStatus);
    }

    return Future<List<DownloadStatus>>.value(downloadStatusList);
  }

  Future<dynamic> _handlePlatformMethodCall(MethodCall call) async {
    switch (call.method.toString()) {
      case "onQueued":
        {
          String mediaId = call.arguments['mediaId'];
          _currentMediaIds.add(mediaId);
          Map<Object?, Object?> downloadStatusMap =
              call.arguments['downloadStatus'];
          DownloadStatus downloadStatus =
              DownloadStatus.fromMap(downloadStatusMap);
          _onQueue(mediaId, downloadStatus);
        }
        break;
      case "onChanged":
        {
          String mediaId = call.arguments['mediaId'];
          Map<Object?, Object?> downloadStatusMap =
              call.arguments['downloadStatus'];
          DownloadStatus downloadStatus =
              DownloadStatus.fromMap(downloadStatusMap);
          if (_currentMediaIds.contains(mediaId)) {
            for (EventListener eventListener in _downloadEventListeners) {
              eventListener.onChanged(mediaId, downloadStatus);
            }
          } else {
            _currentMediaIds.add(mediaId);
            _onQueue(mediaId, downloadStatus);
          }
        }
        break;
      case "onCompleted":
        {
          String mediaId = call.arguments['mediaId'];
          Map<Object?, Object?> downloadStatusMap =
              call.arguments['downloadStatus'];
          DownloadStatus downloadStatus =
              DownloadStatus.fromMap(downloadStatusMap);
          for (EventListener eventListener in _downloadEventListeners) {
            eventListener.onCompleted(mediaId, downloadStatus);
          }
        }
        break;
      case "onFailed":
        {
          String mediaId = call.arguments['mediaId'];
          Map<Object?, Object?> downloadStatusMap =
              call.arguments['downloadStatus'];
          DownloadStatus downloadStatus =
              DownloadStatus.fromMap(downloadStatusMap);
          for (EventListener eventListener in _downloadEventListeners) {
            eventListener.onFailed(mediaId, downloadStatus);
          }
        }
        break;
      case "onDeleted":
        {
          String mediaId = call.arguments['mediaId'];
          _currentMediaIds.remove(mediaId);
          for (EventListener eventListener in _downloadEventListeners) {
            eventListener.onDeleted(mediaId);
          }
        }
    }
  }

  _onQueue(String mediaId, DownloadStatus downloadStatus) {
    for (EventListener eventListener in _downloadEventListeners) {
      eventListener.onQueued(mediaId, downloadStatus);
    }
  }

  /// Add a listener for download events.
  void addDownloadEventListener(EventListener eventListener) {
    _downloadEventListeners.add(eventListener);
    if (!_isDownloadEventListenerAdded && _downloadEventListeners.isNotEmpty) {
      _methodChannel.invokeMethod("addEventListener");
      _isDownloadEventListenerAdded = true;
    }
  }

  /// Remove a listener of download events.
  void removeDownloadEventListener(EventListener eventListener) {
    _downloadEventListeners.remove(eventListener);
    if (_isDownloadEventListenerAdded && _downloadEventListeners.isEmpty) {
      _methodChannel.invokeMethod("removeEventListener");
      _isDownloadEventListenerAdded = false;
    }
  }
}
