import 'dart:async';
import 'dart:io';

import 'package:flutter/foundation.dart';
import 'package:flutter/gestures.dart';
import 'package:flutter/rendering.dart';
import 'package:flutter/services.dart';
import 'package:flutter/material.dart';
import '../controllers/controllers.dart';
import '../types/types.dart';

/// Widget for displaying video from your VdoCipher dashboard.
///
/// On being created, this widget creates a native platform video player and
/// provides a [VdoPlayerController] to control playback. The
/// [VdoPlayerController] reference is returned in the [onPlayerCreated] function
/// provided as an argument in the constructor.
class VdoPlatformView extends StatefulWidget {
  final Function(VdoPlayerController controller) onPlayerCreated;
  final Function(bool isLoaded) onPlayerLoaded;
  final Function(bool isFullscreen)? onFullscreenChange;
  final bool controls;
  final Function(VdoError vdoError) onError;

  const VdoPlatformView(
      {Key? key,
      this.controls = true,
      this.onFullscreenChange,
      required this.onPlayerCreated,
      required this.onPlayerLoaded,
      required this.onError})
      : super(key: key);

  @override
  State<StatefulWidget> createState() => _VdoPlatformViewState();
}

class _VdoPlatformViewState extends State<VdoPlatformView> {
  final String viewType = 'com.vdocipher.flutter/vdoplayerview';

  final Completer<VdoPlayerController> _controller =
      Completer<VdoPlayerController>();
  MethodChannel? _channel;

  @override
  Widget build(BuildContext context) {
    return SizedBox(
        width: double.infinity,
        height: double.infinity,
        child: Platform.isIOS ? _getIOSView() : _getAndroidView());
  }

  @override
  void dispose() async {
    super.dispose();

    // super.dispose() called before async calls because the platform checks for
    // super call immediately after the method returns (void, or a Future<void)
    VdoPlayerController controller = await _controller.future;
    controller.dispose();
  }

  _getAndroidView() {
    return PlatformViewLink(
      viewType: viewType,
      surfaceFactory:
          (BuildContext context, PlatformViewController controller) {
        return AndroidViewSurface(
          controller: controller as AndroidViewController,
          gestureRecognizers: const <Factory<OneSequenceGestureRecognizer>>{},
          hitTestBehavior: PlatformViewHitTestBehavior.opaque,
        );
      },
      onCreatePlatformView: (PlatformViewCreationParams params) {
        return PlatformViewsService.initExpensiveAndroidView(
          id: params.id,
          viewType: viewType,
          layoutDirection: TextDirection.ltr,
          creationParams: {},
          creationParamsCodec: const StandardMessageCodec(),
        )
          ..addOnPlatformViewCreatedListener(
              _onPlatformViewCreated(params.onPlatformViewCreated))
          ..create();
      },
    );
  }

  _getIOSView() {
    return UiKitView(
      viewType: viewType,
      layoutDirection: TextDirection.ltr,
      creationParams: {'buildWithControls': widget.controls},
      creationParamsCodec: const StandardMessageCodec(),
      onPlatformViewCreated: _onIOSPlatformViewCreated,
    );
  }

  _onIOSPlatformViewCreated(int id) {
    _channel = MethodChannel('com.vdocipher.flutter.vdoplayerview_$id');
    _channel!
        .setMethodCallHandler((call) => _handlePlatformMethodCall(call, id));
  }

  _onPlatformViewCreated(Function platformViewCreatedCallback) {
    return ((int id) {
      _channel = MethodChannel('com.vdocipher.flutter.vdoplayerview_$id');
      _channel!
          .setMethodCallHandler((call) => _handlePlatformMethodCall(call, id));
      platformViewCreatedCallback(id);
    });
  }

  Future<dynamic> _handlePlatformMethodCall(MethodCall call, int id) async {
    switch (call.method) {
      case 'onPlatformPlayerCreated':
        setState(() {
          final VdoPlayerController vdoPlayerController = VdoPlayerController(
              _channel,
              EventChannel('com.vdocipher.flutter.vdoplayerview.events_$id'));
          _controller.complete(vdoPlayerController);
          vdoPlayerController.init();
          widget.onPlayerCreated(vdoPlayerController);
        });
        break;
      case "onLoading":
        widget.onPlayerLoaded(false);
        break;
      case "onLoaded":
        widget.onPlayerLoaded(true);
        break;
      case "onLoadError":
        widget.onPlayerLoaded(false);
        Map<Object?, Object?> errorDescription =
            call.arguments['errorDescription'];
        VdoError vdoError = VdoError.fromMap(errorDescription);
        widget.onError.call(vdoError);
        break;
      case "onError":
        widget.onPlayerLoaded(false);
        Map<Object?, Object?> errorDescription =
            call.arguments['errorDescription'];
        VdoError vdoError = VdoError.fromMap(errorDescription);
        widget.onError.call(vdoError);
        break;
      case "onFullscreenChange":
        bool isFullscreen = call.arguments['isFullscreen'] as bool;
        if (widget.onFullscreenChange != null) {
          widget.onFullscreenChange!(isFullscreen);
        }
        break;
      default:
        throw MissingPluginException();
    }
  }
}
