import 'dart:io';

import 'package:flutter/material.dart';
import 'package:vdocipher_flutter/src/views/vdo_styled_player.dart';
import 'package:vdocipher_flutter/vdocipher_flutter.dart';

/// Widget for displaying video from your VdoCipher dashboard along with basic
/// video UI controls.
///
/// On being created, this widget creates a native platform video player and
/// provides a [VdoPlayerController] to control playback. The
/// [VdoPlayerController] reference is returned in the [onPlayerCreated] function
/// provided as an argument in the constructor.
class VdoPlayer extends StatefulWidget {
  final Function(VdoPlayerController controller) onPlayerCreated;
  final Function(bool isFullscreen)? onFullscreenChange;
  final Function(VdoError vdoError) onError;
  final EmbedInfo embedInfo;
  final double aspectRatio;
  final bool controls;
  final Function(bool isInPictureInPictureMode)? onPictureInPictureModeChanged;

  const VdoPlayer(
      {Key? key,
      required this.embedInfo,
      required this.onPlayerCreated,
      this.aspectRatio = 16 / 9,
      this.controls = true,
      this.onFullscreenChange,
      this.onPictureInPictureModeChanged,
      required this.onError})
      : super(key: key);

  @override
  VdoPlayerState createState() => VdoPlayerState();
}

class VdoPlayerState extends State<VdoPlayer> {
  VdoPlayerController? _controller;
  VdoPlatformView? _vdoPlayerView;
  bool _isFullScreen = false;
  bool _isLoaded = false;

  _getVdoPlayerViewWidget() {
    _vdoPlayerView ??= VdoPlatformView(
      key: UniqueKey(),
      controls: widget.controls,
      onFullscreenChange: _onFullscreenChange,
      onError: widget.onError,
      onPlayerCreated: (controller) => _onPlayerCreated(controller),
      onPlayerLoaded: (isLoaded) => _onPlayerLoaded(isLoaded),
    );

    return _vdoPlayerView;
  }

  _onFullscreenChange(isFullscreen) {
    setState(() {
      _isFullScreen = isFullscreen;
    });

    if (widget.onFullscreenChange != null) {
      widget.onFullscreenChange!(isFullscreen);
    }
  }

  _onPlayerLoaded(bool isLoaded) {
    setState(() {
      _isLoaded = isLoaded;
    });
  }

  Stack _buildPlayerWithControlsForIOS() {
    return Stack(
      children: <Widget>[
        Positioned.fill(
          child: _getVdoPlayerViewWidget(),
        ),
        _isLoaded || Platform.isIOS
            ? const SizedBox.shrink()
            : Positioned.fill(
                child: Container(
                  color: Colors.black,
                ),
              ),
      ],
    );
  }

  Widget _buildPlayerWithControlsForAndroid(Size screenSize) {
    return SizedBox(
      width: screenSize.width,
      height: screenSize.height,
      child: VdoStyledPlayer(
        embedInfo: widget.embedInfo,
        aspectRatio: widget.aspectRatio,
        onFullscreenChange: _onFullscreenChange,
        onPlayerCreated: _onPlayerCreated,
        controls: widget.controls,
        onError: widget.onError,
        onPictureInPictureModeChanged: widget.onPictureInPictureModeChanged,
      ),
    );
  }

  @override
  Widget build(BuildContext context) {
    Size screenSize = MediaQuery.of(context).size;

    return SafeArea(
        top: !_isFullScreen,
        left: !_isFullScreen,
        bottom: !_isFullScreen,
        right: !_isFullScreen,
        child: Container(
            width: screenSize.width,
            height: screenSize.height,
            color: Colors.black,
            child: AspectRatio(
                aspectRatio: widget.aspectRatio,
                child: Platform.isIOS
                    ? _buildPlayerWithControlsForIOS()
                    : _buildPlayerWithControlsForAndroid(screenSize))));
  }

  _onPlayerCreated(VdoPlayerController controller) {
    _controller = controller;
    _controller!.load(widget.embedInfo);
    widget.onPlayerCreated(controller);
  }
}
